#   DreamBot - An Experimental IRC Bot written in Perl.
#   Copyright (C) 1999-2002  Osman "spectre" Keskin
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

sub parser
{
  my ($line) = $_[0];
  if ($line =~ /^ERROR :Closing Link: (.*)/)
  {
    console("*** ERROR: $1\n");
    close($sock);
    serverconnect();
    return;
  }
  elsif ($line =~ /^PING [:]?(.*)/)
  {
    sockwrite($sock, "PONG :$1");
  }
  elsif ($line =~ /^:([^: ]*)!([^: ]*) PRIVMSG ([^: ]+) :(.*)$/)
  {
    $unick = $1;
    $uhost = $2;
    $uchan = $3;
    $utext = $4;
    $uhost =~ /(.*)\@(.*)/;
    $uident = $1;
    $uaddy  = $2;
    $query  = false;
    if (lc($uchan) eq lc($nick)) { $uchan = $unick; $query = true; }
    $nnick = checkhost(lc($unick), $uhost);

    if ($utext =~ /^\x01PING (.*)\x01/)
    {
      console("[$unick] PING\n");
      sockwrite($sock, "NOTICE $unick :\x01PING $1\x01");
      foreach $logfile (values %logfiles) { chanlog($logfile, "[$unick] PING\n"); }
    }
    elsif ($utext =~ /^\x01VERSION\x01/)
    {
      console("[$unick] VERSION\n");
      sockwrite($sock, "NOTICE $uchan :\x01VERSION $version\x01");
      foreach $logfile (values %logfiles) { chanlog($logfile, "[$unick] VERSION\n"); }
    }
    elsif ($utext =~ /^\x01ACTION (.*)\x01/)
    {
      console("[$uchan] $unick $1\n");
      if ($logging{lc($uchan)} eq true) { chanlog($logfiles{lc($uchan)}, "* $unick $1\n"); }
    }
    else
    {
      console("\[$uchan\] <$unick> $utext\n");
      if ($logging{lc($uchan)} eq true) { chanlog($logfiles{lc($uchan)}, "<$unick> $utext\n"); }
      if ($query ne true)
      {
        if ($nnick ne false) { $tnick = $nnick; } else { $tnick = $unick; }
        $talkers{lc($tnick)} += length($utext);
      }
      if (defined($on_msg_handler)) { &$on_msg_handler($uchan, $nnick, $unick, $uident, $uaddy, $utext); }

      if ($learning{lc($uchan)} eq true)
      {
        @sentences = split (/\. /, $utext);
        foreach $sentence (@sentences)
        {
          if ($sentence =~ /^(?i:$nick)?[,:\-\; ]*(.*)(?: is| are|\'s|\'re) (.*)/i ||
            $sentence =~ /^(?i:$nick)?[,:\-\; ]*(.*)(?: =| ->) (.*)/i)
          {
            $key = lc($1);
            $val = $2;
            if ($key !~ /^(what|that|it|these|those|he|she|who|where|how)$/i && $nick ne $unick && !defined($facts{$key}))
            {
              if (substr($val, 0, 7) ne "<reply>")
              {
                $val =~ s/your/my/gi;
                $val =~ s/you/I/gi;
              }
              $key =~ s/your/my/gi;
              $key =~ s/you/I/gi;
              console("*** learning: $key -> $val\n");
              open(FILE, ">>$datadir/facts");
              printf FILE "%s: %s -> %s\n", $unick, $key, $val;
              close(FILE);
              $facts{$key} = "$unick: $val";
            }
          }
        }
      }
    }
    if ($utext =~ /^(?i:$nick)[,:\-\; ]*how (are|r) (you|ya|u)[ \?!]*/i)
    {
      privmsg($uchan, "I'm ok, $unick.");
    }
    elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*(?:fm|fresh|freshmeat)(?: news| headlines)?[ \?!]*$/i)
    {
      if ($public_fm{lc($uchan)} ne true && $query ne true)
      {
        privmsg($uchan, "$unick, use query for freshmeat");
      }
      elsif (!fork())
      {
        $data = httpget("www.freshmeat.net", 80, "/backend/fm.rdf");
        if ($data ne "")
        {
          $ndata = backend($data);
          privmsg($uchan, "Freshmeat: $ndata");
        }
        else
        {
          privmsg($uchan, "$unick, an error occured... (that's all i can say)");
        }
        exit(0);
      }
    }
    elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*(?:\/\.|slash|slashdot)(?: news| headlines)?[ \?!]*$/i)
    {
      if ($public_sd{lc($uchan)} ne true && $query ne true)
      {
        privmsg($uchan, "$unick, use query for slashdot");
      }
      elsif (!fork())
      {
        $data = httpget("www.slashdot.org", 80, "/slashdot.rdf");
        if ($data ne "")
        {
          $ndata = backend($data);
          privmsg($uchan, "Slashdot: $ndata");
        }
        else
        {
          privmsg($uchan, "$unick, an error occured... (that's all i can say)");
        }
        exit(0);
      }
    }
    elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*(?:translate|xlat|xlate|xl) (to|from) ([^ :;,\-]*)[ :;,\-]*(.*)$/i)
    {
      if (!fork())
      {
        if ($1 eq "to")
        {
          $lang = "en_" . $2;
        }
        else
        {
          $lang = $2 . "_en";
        }
        $string = urlencode($3);
        undef($data);
        $data = httppost("babel.altavista.com", 80, "/tr", "doit=done&tt=urltext&urltext=$string&url=http%3A%2F%2F&lp=$lang");
        if ($data ne "" && ($data =~ /<textarea rows=\"3\" wrap\=virtual cols\=\"56\" name\=\"q\">([^<>]*)\n<\/textarea>/ || $data =~ /<td bgcolor\=white>([^<>]*)\n<\/td>\n/))
        {
          $xlat = $1;
          $xlat =~ s/\n/ /g;
          privmsg($uchan, "$unick, translation: $xlat");
        }
        else
        {
          privmsg($uchan, "$unick, sorry, an error occured... (yes. that's all i can say)");
        }
        exit(0);
      }
    }
    elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*(?:pass|passwd|password) (.+)$/i && $query eq true)
    {
      if ($nnick eq false)
      {
        privmsg($uchan, "sorry, you're not on my user list");
      }
      else
      {
        $password{$nnick} = crypt($1, substr($1, 0, 2));
        privmsg($uchan, "your password is set");
        savepasswords();
      }
    }
    elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*op (.+)$/i && $query eq true)
    {
      if ($password{$nnick} eq crypt($1, substr($1, 0, 2)))
      {
        foreach $key (keys %channels)
        {
          if ($uflags{$nnick} =~ /o/ && ($uchans{$nnick} =~ /$key/i || $uchans{$nnick} eq "*")) { sockwrite($sock, "MODE $key +o $unick"); }
        }
      }
      else
      {
        privmsg($uchan, "wrong password");
      }
    }

    if ($utext =~ /^(?i:$nick)?[,:\-\; ]*\./ && $nnick ne false && $uflags{$nnick} =~ /m/)
    {
      if ($utext =~ /^(?i:$nick)?[,:\-\; ]*.(?:raw|order|quote) (.+)/i)
      {
        sockwrite($sock, "$1");
        console("*** order: $1\n");
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.(?:add|set)user ([^ ]+) ([^\[\]\(\)\+ ]*) ([^ ]*) (.*)$/i)
      {
        $user = $1;
        $host = $2;
        $flag = $3;
        $chan = $4;
        $host =~ s/\*/\.\*/g;
        $host =~ s/\?/\.\?/g;
        $host =~ s/,/ /g;
        $chan =~ s/,/ /g;
        $uhosts{lc($user)} = $host;
        $uflags{lc($user)} = $flag;
        $uchans{lc($user)} = $chan;
        console "*** added user $user\n";
        privmsg($uchan, "added user $user");
        saveusers();
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.remuser (.+)$/i)
      {
        console "*** removed user $1\n";
        privmsg($uchan, "removed user $1");
        delete $uhosts{lc($1)};
        delete $uflags{lc($1)};
        delete $uchans{lc($1)};
        saveusers();
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.addchan ([^ :]+)[ ]?(.*)$/i)
      {
        $chan = lc($1);
        $chanflags{$chan} = $2;
        processchanflags($chan, $2);
        privmsg($uchan,         "added channel $chan");
        $channels{$chan} =~ /([^,]*)[,]?(.*)/;
        sockwrite($sock, "JOIN $chan $2");
        console("*** added channel: $chan\n");
        savechannels();
        savetopiclocks();
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.remchan (.+)$/i)
      {
        if (defined($channels{lc($1)}))
        {
          delete $channels{lc($1)};
          privmsg($uchan,  "removed channel $1");
          sockwrite($sock, "PART $1");
          console("*** removed channel: $1\n");
          savechannels();
        }
        else
        {
          privmsg($uchan, "no such channel: $1");
        }
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.showuser (.+)$/i)
      {
        $user = $1;
        console "*** showing user $user\n";
        if (defined($uhosts{lc($user)}))
        {
          $host = $uhosts{lc($user)};
          $host =~ s/\.\*/\*/g;
          $host =~ s/\.\?/\?/g;
          $host =~ s/ /,/g;
          $chan = $uchans{lc($user)};
          $chan =~ s/ /,/g;
          privmsg($uchan, "[$user] hosts: [$host] flags: [$uflags{lc($user)}] chans: [$chan]");
        }
        else
        {
          privmsg($uchan, "I don't know any person named $1");
        }
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.save$/i)
      {
        savechannels();
        saveusers();
        savefacts();
        savenotes();
        saveseen();
        savetopiclocks();
        savetalkers();
        savepasswords();
        privmsg($uchan, "saved all");
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.load$/ || $utext =~ /^(?i:$nick)?[,:\-\; ]*.rehash$/i)
      {
        undef %channels;
        undef %uhosts;
        undef %uflags;
        undef %uchans;
        undef %facts;
        undef %notes;
        undef %topiclock;
        loadchannels();
        loadusers();
        loadfacts();
        loadnotes();
        loadtopiclocks();
        privmsg($uchan, "reloaded all");
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.topiclock (.+)$/i)
      {
        $key = $1;
        undef($nchan);
        if ($query eq true)
        {
          if ($utext =~ /^(?i:$nick)?[,:\-\; ]*.topiclock (.+) (.+)$/i)
          {
            $key   = $1;
            $nchan = $2;
          }
          else
          {
            privmsg($uchan, "Use .topiclock <on/off> <channel> syntax on query");
          }
        }
        else
        {
          $nchan = $uchan;
        }
        if (defined($nchan))
        {
          if ($key eq "on")
          {
            $topiclock{lc($nchan)} = true;
            undef($oldtopic{lc($nchan)});
            sockwrite($sock, "TOPIC $nchan");
            console("*** topiclock on $nchan is on\n");
            privmsg($uchan, "topiclock on $nchan is on");
          }
          else
          {
            $topiclock{lc($nchan)} = false;
            console("*** topiclock on $nchan is off\n");
            privmsg($uchan, "topiclock on $nchan is off");
          }
        }
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.modelock (.+)$/i)
      {
        $key = $1;
        undef($nchan);
        if ($query eq true)
        {
          if ($utext =~ /^(?i:$nick)?[,:\-\; ]*.modelock (.+) (.+)$/i)
          {
            $key   = $1;
            $nchan = $2;
          }
          else
          {
            privmsg($uchan, "Use .modelock <on/off> <channel> syntax on query");
          }
        }
        else
        {
          $nchan = $uchan;
        }
        if (defined($nchan))
        {
          if ($key eq "on")
          {
            $modelock{lc($nchan)} = true;
            sockwrite($sock, "MODE $nchan");
            console("*** modelock on $nchan is on\n");
            privmsg($uchan, "modelock on $nchan is on");
          }
          else
          {
            $modelock{lc($nchan)} = false;
            console("*** modelock on $nchan is off\n");
            privmsg($uchan, "modelock on $nchan is off");
          }
        }
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.bitch(?:mode)? (.+)$/i)
      {
        $key = $1;
        undef($nchan);
        if ($query eq true)
        {
          if ($utext =~ /^(?i:$nick)?[,:\-\; ]*.bitchmode (.+) (.+)$/i)
          {
            $key   = $1;
            $nchan = $2;
          }
          else
          {
            privmsg($uchan, "Use .bitchmode <on/off> <channel> syntax on query");
          }
        }
        else
        {
          $nchan = $uchan;
        }
        if (defined($nchan))
        {
          if ($key eq "on")
          {
            $bitchmode{lc($nchan)} = true;
            console("*** bitchmode on $nchan is on\n");
            privmsg($uchan, "bitchmode on $nchan is on");
          }
          else
          {
            $bitchmode{lc($nchan)} = false;
            console("*** bitchmode on $nchan is off\n");
            privmsg($uchan, "bitchmode on $nchan is off");
          }
        }
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.opprot (.+)$/i)
      {
        $key = $1;
        undef($nchan);
        if ($query eq true)
        {
          if ($utext =~ /^(?i:$nick)?[,:\-\; ]*.opprot (.+) (.+)$/i)
          {
            $key   = $1;
            $nchan = $2;
          }
          else
          {
            privmsg($uchan, "Use .opprot <on/off> <channel> syntax on query");
          }
        }
        else
        {
          $nchan = $uchan;
        }
        if (defined($nchan))
        {
          if ($key eq "on")
          {
            $opprot{lc($nchan)} = true;
            console("*** op protection on $nchan is on\n");
            privmsg($uchan, "op protection on $nchan is on");
          }
          else
          {
            $opprot{lc($nchan)} = false;
            console("*** op protection on $nchan is off\n");
            privmsg($uchan, "op protection on $nchan is off");
          }
        }
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.learning (.+)$/i)
      {
        $key = $1;
        undef($nchan);
        if ($query eq true)
        {
          if ($utext =~ /^(?i:$nick)?[,:\-\; ]*.learning (.+) (.+)$/i)
          {
            $key   = $1;
            $nchan = $2;
          }
          else
          {
            privmsg($uchan, "Use .learning <on/off> <channel> syntax on query");
          }
        }
        else
        {
          $nchan = $uchan;
        }
        if (defined($nchan))
        {
          if ($key eq "on")
          {
            $learning{lc($nchan)} = true;
            console("*** learning on $nchan is on\n");
            privmsg($uchan, "learning on $nchan is on");
          }
          else
          {
            $learning{lc($nchan)} = false;
            console("*** learning on $nchan is off\n");
            privmsg($uchan, "learning on $nchan is off");
          }
        }
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.log (.+)$/i)
      {
        $key = $1;
        undef($nchan);
        if ($query eq true)
        {
          if ($utext =~ /^(?i:$nick)?[,:\-\; ]*.log (.+) (.+)$/i)
          {
            $key   = $1;
            $nchan = $2;
          }
          else
          {
            privmsg($uchan, "Use .log <on/off> <channel> syntax on query");
          }
        }
        else
        {
          $nchan = $uchan;
        }
        if (defined($nchan))
        {
          if ($key eq "on")
          {
            $logging{lc($nchan)}  = true;
            $logfiles{lc($nchan)} = openlogfile(lc($nchan));
            console("*** logging on $nchan is on\n");
            privmsg($uchan, "logging on $nchan is on");
          }
          else
          {
            $logging{lc($nchan)} = false;
            close($logfiles{lc($nchan)});
            delete($logfiles{lc($nchan)});
            console("*** logging on $nchan is off\n");
            privmsg($uchan, "logging on $nchan is off");
          }
        }
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.chaninfo[ ]?(.*)$/i)
      {
        if ($query eq true)
        {
          if ($1 eq "") { privmsg($uchan, "Use .chaninfo <channel> syntax on query"); }
          else { privmsg($nchan, "log[$logging{lc($nchan)}], topiclock[$topiclock{$nchan}], modelock[$modelock{$nchan}], bitchmode[$bitchmode{$nchan}], op protection[$opprot{$nchan}], freshmeat[$public_fm{$nchan}], slashdot[$public_sd{$nchan}], learning[$learning{$nchan}]"); }
        }
        else
        {
          if ($1 ne "") { $nchan = lc($1); } else { $nchan = lc($uchan); }
          privmsg($nchan, "log[$logging{lc($nchan)}], topiclock[$topiclock{$nchan}], modelock[$modelock{$nchan}], bitchmode[$bitchmode{$nchan}], op protection[$opprot{$nchan}], freshmeat[$public_fm{$nchan}], slashdot[$public_sd{$nchan}], learning[$learning{$nchan}]");
        }
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.nick (.+)$/i)
      {
        $nick = quotemeta($1);
        sockwrite($sock, "NICK :$1");
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.join ([^ ]+)[ ]?(.*)$/i)
      {
        privmsg($uchan,  "joining $1");
        sockwrite($sock, "JOIN $1 $2");
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.part (.+)$/i)
      {
        privmsg($uchan,  "leaving $1");
        sockwrite($sock, "PART $1");
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.mode ([^ ]*) (.+)$/i)
      {
        sockwrite($sock, "MODE $1 $2");
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.op ([^ ]*)[ ]?([^ ]*)$/i)
      {
        if ($query eq false)
        {
          sockwrite($sock, "MODE $uchan +o $1");
        }
        elsif ($2 eq "")
        {
          privmsg($uchan, "You need to use .op <person> <channel> syntax on query.");
        }
        else
        {
          sockwrite($sock, "MODE $2 +o $1");
        }
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.d[e]?op ([^ ]*)[ ]?([^ ]*)$/i)
      {
        if ($query eq false)
        {
          sockwrite($sock, "MODE $uchan -o $1");
        }
        elsif ($2 eq "")
        {
          privmsg($uchan, "You need to use .d[e]op <person> <channel> syntax on query.");
        }
        else
        {
          sockwrite($sock, "MODE $2 -o $1");
        }
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.flag[s]? (.*) (.*)$/i)
      {
        if (!defined($uhosts{lc($1)}))
        { privmsg($uchan, "I don't know any person named $1"); }
        else
        {
          $unick = $1;
          $flags = $2;
          if ($flags =~ /^\+(.*)/)
          {
            $uflags{lc($unick)} .= $1;
          }
          elsif ($flags =~ /^\-(.*)/)
          {
            $newflags = $1;
            @nflags = split (//, $newflags);
            foreach $flag (@nflags)
            {
              $uflags{lc($unick)} =~ s/$flag/\-/g;
            }
            $newflags = "";
            @flaglist = split (//, $uflags{lc($unick)});
            foreach $flag (@flaglist)
            {
              if ($flag ne '-') { $newflags .= $flag; }
            }
            $uflags{lc($unick)} = $newflags;
          }
          else
          {
            $uflags{lc($unick)} = $flags;
          }
          console("*** $unick\'s flags are now: $uflags{lc($unick)}\n");
          privmsg($uchan, "$unick\'s flags are now: $uflags{lc($unick)}");
        }
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.addhost (.*) ([^\[\]\(\)\+]*)$/i)
      {
        if (!defined($uhosts{lc($1)})) { privmsg($uchan, "I don't know any person named $1"); }
        else
        {
          $user     = $1;
          $host     = $2;
          $orighost = $2;
          $host =~ s/\*/\.\*/g;
          $host =~ s/\?/\.\?/g;
          $uhosts{lc($user)} .= " $host";
          privmsg($uchan, "added host $orighost to $user");
        }
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.remhost ([^ ]*) ([^ ]*)$/i)
      {
        if (!defined($uhosts{lc($1)})) { privmsg($uchan, "I don't know any person named $1"); }
        else
        {
          $user     = $1;
          $host     = $2;
          $orighost = $2;
          $host =~ s/\*/\.\*/g;
          $host =~ s/\?/\.\?/g;
          @hosts = split (' ', $uhosts{lc($user)});
          undef($newhosts);
          $changed = 0;

          for ($i = 0; $i <= $#hosts; $i++)
          {
            if ($hosts[$i] ne $host) { $newhosts .= "$hosts[$i] "; }
            else { $changed = 1; }
          }
          if ($changed == 0) { privmsg($uchan, "host not found"); }
          else { chop($newhosts); $uhosts{lc($user)} = $newhosts; privmsg($uchan, "host $orighost removed from $user"); }
        }
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.addchanuser (.*) (.*)$/i)
      {
        if (!defined($uhosts{lc($1)})) { privmsg($uchan, "No such user: $1"); }
        else { $uchans{lc($1)} .= " $2"; privmsg($uchan, "$1's channels are now: $uchans{lc($1)}"); }
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.die[ ]?(.*)$/i)
      {
        console("*** die by $unick\n");
        if ($1 ne "") { $reason = $1; } else { $reason = ".die by $unick"; }
        sockwrite($sock, "QUIT :$reason");
        close($sock);
        unlink("$ENV{HOME}/.dreambot.pid");
        if ($log eq true)
        {
          printf LOGFILE "[end]\n";
          close(LOGFILE);
        }
        exit(0);
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.restart[ ]?(.*)$/i)
      {
        console("*** restart by $unick\n");
        if ($1 ne "") { $reason = $1; } else { $reason = ".restart by $unick"; }
        sockwrite($sock, "QUIT :$reason");
        close($sock);
        unlink("$ENV{HOME}/.dreambot.pid");
        if ($log eq true)
        {
          printf LOGFILE "[end]\n";
          close(LOGFILE);
        }
        exec("$0$params");
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.ctcp ([^ ]*) (.*)$/i)
      {
        console("*** ctcp: $2\n");
        sockwrite($sock, "PRIVMSG $1 :\x01$2\x01");
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.jump (.*)[:]?(.*)$/i)
      {
        $server = $1;
        if ($2 eq "") { $port = 6667; } else { $port = $2; }
        sockwrite($sock, "QUIT :.jump by $unick");
        close($sock);
        jumptoserver();
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.cycle ([^ ]*)$/i)
      {
        privmsg($uchan, "cycling $1");
        $chan = $1;
        $channels{$chan} =~ /[^,]*[,]?(.*)/;
        sockwrite($sock, "PART $chan");
        sockwrite($sock, "JOIN $chan $1");
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.userlist$/i)
      {
        $i = 0; $total = 0; $data = "";
        foreach $key (sort keys %uhosts)
        {
          $data .= "$key";
          $i++; $total++;
          if ($i == 20) { $i = 0; sockwrite($sock, "$data"); $data = "PRIVMSG $uchan :"; }
          else { $data .= ","; }
        }
        if (length($data) > 16) { chop($data); privmsg($uchan, "$data"); }
        privmsg($uchan, "Total $total users");
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.chanlist$/i)
      {
        undef($data);
        foreach $key (sort keys %channels)
        {
          $data .= "$key, ";
        }
        chop($data); chop($data);
        privmsg($uchan, "I'm on $data");
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.serverlist$/i)
      {
        $data = "server list: ";
        foreach $key (@servers)
        {
          $data .= "$key,";
        }
        chop($data);
        privmsg($uchan, $data);
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.(?:say|msg) ([^ ]*) (.+)$/i)
      {
        privmsg($1, $2);
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.me ([^ ]*) (.+)$/i)
      {
        privmsg($1, "\x01ACTION $2\x01");
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.topic ([^ ]*) (.*)$/i)
      {
        sockwrite($sock, "TOPIC $1 :$2");
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.kick ([^ ]*) ([^ ]*) (.+)$/i)
      {
        sockwrite($sock, "KICK $1 $2 :$3");
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.ban ([^ ]*) (.+)$/i)
      {
        sockwrite($sock, "MODE $1 +b $2");
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.unban ([^ ]*) (.+)$/i)
      {
        sockwrite($sock, "MODE $1 -b $2");
      }
      elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*.help[ ]?(.*)$/i)
      {
        if ($1 eq "")
        {
          privmsg($uchan, "available commands are: raw|order|quote ctcp cycle adduser|setuser remuser showuser serverlist userlist chanlist load|rehash save topiclock modelock bitchmode opprot log nick join part say topic mode op d[e]op kick ban unban flag[s] addhost remhost addchan remchan chaninfo addchanuser die restart jump help");
        }
        elsif (defined($help{lc($1)}))
        {
          @helptext = split (/\n/, $help{lc($1)});
          foreach $helpline (@helptext)
          {
            privmsg($uchan, "$helpline");
          }
        }
      }
    }

    if ($utext =~ /^(?i:$nick)[,:\-\; ]*(?:note|memo|message|msg)(?: to)? ([^,:\; ]*)[,:\; ]*(.*)$/i)
    {
      if (!defined($uhosts{lc($1)}))
      {
        privmsg($uchan, "$unick, I don't know any person named $1");
      }
      else
      {
        ($sec, $min, $hour, $day, $mon, $year, $wday, $yday, $isdst) = localtime(time());
        $text = sprintf("%02d/%02d/%2d %02d:%02d:%02d", $day, $mon, $year, $hour, $min, $sec);
        $notes{lc($1)} .= "$unick: $2 [$text]\x01";
        privmsg($uchan, "$unick, your note is saved");
      }
    }
    elsif ($utext =~ /^read notes$/i)
    {
      if (defined($notes{$nnick}))
      {
        @nts = split ("\x01", $notes{$nnick});
        for ($i = 0; $i <= $#nts; $i++)
        {
          $nts[$i] =~ /^([^:]*): (.*)$/;
          privmsg($uchan, "[$i] $1: $2");
        }
      }
      else
      {
        privmsg($uchan, "You've no notes waiting.");
      }
    }
    elsif ($utext =~ /^delete (notes|note)[ ]?(.*)$/i)
    {
      if ($1 eq "notes")
      {
        delete $notes{$nnick};
        privmsg($uchan, "Your notes are deleted.");
      }
      else
      {
        @nts = split ("\x01", $notes{$nnick});
        if ($2 > $#nts)
        {
          privmsg($uchan, "Invalid note number");
        }
        else
        {
          undef($newnotes);
          for ($i = 0; $i <= $#nts; $i++)
          {
            if ($i != $2) { $newnotes .= "$nts[$i]\n"; }
          }
          $notes{$nnick} = $newnotes;
          privmsg($uchan, "Note #$2 is deleted.");
        }
      }
    }
    elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*seen ([^ \?!\*]*)[ \?!]*$/i)
    {
      $seenkey = lc($1);
      if ($seenkey eq lc($nick))
      {
        privmsg($uchan, "$unick, looking for me?");
      }
      elsif ($seenkey eq lc($unick))
      {
        privmsg($uchan, "$unick, looking for yourself?");
      }
      else
      {
        $seenflag = true;
        $seennick = $unick;
        sockwrite($sock, "NAMES $uchan");
      }
    }
    elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*topten[ ]?([^ \?!]*)[ \?!]*$/i)
    {
      $i = 0; undef($data);
      foreach $key (sort { $talkers{$b} <=> $talkers{$a} } keys %talkers)
      {
        $i++;
        $count = $talkers{$key};
        if ($count > 1000)
        {
          $count /= 1000;
          $text = sprintf("%.2fk", $count);
        }
        else
        {
          $text = $count;
        }
        $data .= "$i. $key\[$text\] ";
        last if ($i == 10);
      }
      chop($data);
      $nchan = $uchan;
      if ($1 ne "") { $nchan = $1; }
      privmsg($nchan, $data);
    }
    elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*count[ ]?([^ \?!]*)[ \?!]*$/i)
    {
      undef($tnick);
      if ($1 ne "")
      {
        if (defined($uhosts{lc($1)}))
        {
          $tnick = $1;
        }
        else
        {
          privmsg($uchan, "I don't know anyone named $1");
        }
      }
      else
      {
        $tnick = $nnick;
      }
      if (defined($tnick))
      {
        privmsg($uchan, "$tnick wrote $talkers{lc($tnick)} letters");
      }
    }
    elsif ($utext =~ /^(?i:$nick)?[,:\-\; ]*place[ ]?([^ \?!]*)[ \?!]*$/i)
    {
      undef($tnick);
      if ($1 ne "")
      {
        if (defined($uhosts{lc($1)}))
        {
          $tnick = $1;
        }
        else
        {
          privmsg($uchan, "I don't know anyone named $1");
        }
      }
      else
      {
        if ($nnick ne false) { $tnick = lc($nnick); }
        else { $tnick = lc($unick); }
      }
      if (defined($talkers{$tnick}))
      {
        $i = 0;
        foreach $key (sort { $talkers{$b} <=> $talkers{$a} } keys %talkers)
        {
          $i++;
          last if ($key eq lc($tnick));
        }
        privmsg($uchan, "$tnick is on $i. place");
      }
    }
    elsif ($utext =~ /^(?i:$nick)[,:\-\; ]*(?:what|what the fuck|what the hell|who|where|how)( is| are|\'s|\'re) ([^\?!]*)[ \?!]*$/i)
    {
      $dum = $1;
      $key = lc($2);
      if    ($dum =~ /\'s/i)  { $dum = " is"; }
      elsif ($dum =~ /\'re/i) { $dum = " are"; }
      if ($key eq "you") { $key = "I"; $dum = " am"; }
      $key =~ s/your/my/gi;
      if (defined($facts{$key}))
      {
        $facts{$key} =~ /^([^:]*): (.*)$/;
        $who = $1;
        $val = $2;
        console("*** object: $key -> $val\n");
        if (substr($val, 0, 7) eq "<reply>")
        {
          $reply = $unick . ":" . substr($val, 7);
        }
        else
        {
          $num = int(rand(6));

          if    ($num == 0) { $reply = "according to $who, $key$dum $val"; }
          elsif ($num == 1) { $reply = "I think $key$dum $val"; }
          elsif ($num == 2) { $reply = "I\'ve heard that $key$dum $val"; }
          elsif ($num == 3) { $reply = "$key$dum $val I guess.."; }
          elsif ($num >= 4) { $reply = "$key$dum $val"; }
        }
        privmsg($uchan, "$reply");
      }
      else
      {
        $reply = pickrandom(@dunno);
        privmsg($uchan, "$unick, $reply");
      }
    }
    elsif (($utext =~ /^(?i:$nick)?[,:\-\; ]*\?\? (.*)$/i) ||
      ($utext =~ /^(?i:$nick)[,:\-\; ]* ([^\?!]*)[ \?!]*\?$/i))
    {
      $key = lc($1);
      $key =~ s/your/my/gi;
      $key =~ s/you/I/gi;
      if (defined($facts{$key}))
      {
        $facts{$key} =~ /^([^:]*): (.*)$/;
        $who = $1;
        $val = $2;
        console("*** object: $key -> $val\n");
        if (substr($val, 0, 7) eq "<reply>")
        {
          $reply = $unick . ":" . substr($val, 7);
        }
        else
        {
          $reply = "[\x02$key\x02]: $val [$who]";
        }
        privmsg($uchan, "$reply");
      }
      else
      {
        $reply = pickrandom(@dunno);
        privmsg($uchan, "$unick, $reply");
      }
    }
    elsif ($utext =~ /^(?i:$nick)[,:\-\; ]*forget (.*)[ .,]*$/i)
    {
      $key = $1;
      $key =~ s/your/my/gi;
      $key =~ s/you/I/gi;
      if ($key ne "I") { $key = lc($key); }
      if (defined($facts{$key}))
      {
        delete $facts{$key};
        privmsg($uchan, "$unick, I forgot $key");
        console("*** forgot $1\n");
        savefacts();
      }
      else
      {
        privmsg($uchan, "$unick, I didn't know anything like that");
      }
    }
  }
  elsif ($line =~ /^:(.*)!(.*) JOIN [:]?(.*)$/)
  {
    $unick = $1;
    $uhost = $2;
    $uchan = $3;
    $uhost =~ /(.*)\@(.*)/;
    $uident = $1;
    $uaddy  = $2;
    $nnick  = checkhost(lc($unick), $uhost);
    console("\[$uchan\] $unick ($uident\@$uaddy) joined\n");
    if ($logging{lc($uchan)} eq true) { chanlog($logfiles{lc($uchan)}, "*** $unick ($uident\@$uaddy) joined\n"); }
    $names{lc($uchan)} .= " $unick";
    $nnick = checkhost(lc($unick), $uhost);
    $nhosts{lc($unick)} = $uident . '@' . $uaddy;

    if ($nnick ne false)
    {
      if ($uchans{$nnick} =~ $uchan || $uchans{$nnick} eq "*" || lc($uchan) eq lc($nick))
      {
        if ($uflags{$nnick} =~ /o/) { sockwrite($sock, "MODE $uchan +o $unick"); }
        elsif ($uflags{$nnick} =~ /v/) { sockwrite($sock, "MODE $uchan +v $unick"); }
        elsif ($uflags{$nnick} =~ /b/) { sockwrite($sock, "MODE $uchan +b *!$uhost\nKICK $uchan $unick :autobanned"); }
      }
      if (defined($notes{$nnick}) && $informed{$nnick} ne true)
      {
        @nts = split ("\x01", $notes{$nnick});
        $count = $#nts + 1;
        if ($count > 1) { $data = "$count notes" } else { $data = "a note"; }
        sockwrite($sock, "NOTICE $unick :You got $data waiting, \"/msg $nick read notes\" to read them, \"/msg $nick delete notes\" to delete them all, or \"/msg $nick delete note <num>\" to delete a specific note.");
        $informed{$nnick} = true;
      }
    }
    if ($unick eq $nick)
    {
      sockwrite($sock, "TOPIC $uchan");
      sockwrite($sock, "MODE $uchan");
    }
    if (defined(@on_join_handler))
    {
      printf("blah\n");
      foreach $func (@on_join_handler)
      {
        &$func($uchan, $nnick, $unick, $uident, $uaddy);
      }
    }
  }
  elsif ($line =~ /^:(.*) PART ([^: ]+)[ :]*(.*)$/)
  {
    $uhost  = $1;
    $uchan  = $2;
    $reason = $3;
    $uhost =~ /(.*)!(.*)\@(.*)/;
    $unick  = $1;
    $uident = $2;
    $uaddy  = $3;
    $nnick  = checkhost(lc($unick), $uhost);
    console("\[$uchan\] $unick ($uident\@$uaddy) left \[$reason\]\n");
    if ($logging{lc($uchan)} eq true) { chanlog($logfiles{lc($uchan)}, "*** $unick ($uident\@$uaddy) left \[$reason\]\n"); }
    @nicklist = split (' ', $names{lc($uchan)}); undef($newlist);

    foreach $key (@nicklist)
    {
      $user = quotemeta($unick);
      if ($key !~ /$user/i) { $newlist .= "$key "; }
    }
    chop($newlist);
    $names{lc($uchan)} = $newlist;
    if ($newlist =~ /^[@\+]?$nick$/i)
    {
      sockwrite($sock, "NAMES $uchan");
    }
    $seen{lc($unick)} = time() . "\x01$uchan\x01$reason\x01leaving";
    delete $nhosts{lc($unick)};
    if (defined($on_part_handler)) { &$on_part_handler($uchan, $nnick, $unick, $uident, $uaddy, $reason); }
  }
  elsif ($line =~ /^:(.*) QUIT :?(.*)$/)
  {
    $uhost  = $1;
    $reason = $2;
    $uhost =~ /(.*)!(.*)\@(.*)/;
    $unick  = $1;
    $uident = $2;
    $uaddy  = $3;
    $nnick  = checkhost(lc($unick), $uhost);
    console("*** $unick ($uident\@$uaddy) signed off \[$reason\]\n");

    $user = quotemeta($unick);
    foreach $chan (keys %channels)
    {
      @nicklist = split (' ', $names{lc($chan)});
      undef($newlist); $found = false;
      foreach $key (@nicklist)
      {
        if ($key !~ /$user/i) { $newlist .= "$key "; } else { $found = true; }
      }
      chop($newlist);
      $names{lc($chan)} = $newlist;
      if ($newlist =~ /^[@\+]?$nick$/i)
      {
        sockwrite($sock, "NAMES $chan");
      }
      if ($found eq true && $logging{lc($chan)} eq true && defined($logfiles{lc($chan)}))
      {
        chanlog($logfiles{lc($chan)}, "*** $unick ($uident\@$uaddy) signed off \[$reason\]\n");
      }
    }
    $seen{lc($unick)} = time() . "\x01$uchan\x01$reason\x01quitting";
    delete $nhosts{lc($unick)};
    if (defined($on_quit_handler)) { &$on_quit_handler($uchan, $nnick, $unick, $uident, $uaddy, $reason); }
  }
  elsif ($line =~ /^:(.*)!(.*)\@(.*) KICK ([^: ]+) (.*) :(.*)$/)
  {
    console("\[$4\] $5 is kicked by $1 \($6\)\n");
    if ($logging{lc($4)} eq true) { chanlog($logfiles{lc($4)}, "*** $5 is kicked by $1 \($6\)\n"); }
    if ($5 eq $nick)
    {
      $chan = $4;
      $channels{$chan} =~ /[^,]*[,]?(.*)/;
      sockwrite($sock, "JOIN $chan $1");
    }
    $seen{lc($5)} = time() . "\x01$4\x01$6\x01being kicked";
    delete $nhosts{lc($5)};
  }
  elsif ($line =~ /^:(.*) MODE ([^: ]+) [:]?([^ ]*)[ ]?(.*)$/)
  {
    $uhost = $1;
    $uchan = $2;
    $modes = $3;
    $ppl   = $4;
    if ($uhost =~ /(.*)!(.*)\@(.*)/)
    {
      $uhost = $1;
    }
    console("\[$uchan\] $uhost sets mode: $modes $ppl\n");
    if ($logging{lc($uchan)} eq true) { chanlog($logfiles{lc($uchan)}, "*** $uhost set mode: $modes $ppl\n"); }

    @cnicks   = split (' ', $ppl);
    @modelist = split (//,  $modes);

    undef($flag); $i = 0; $sendmodes = false;
    foreach $mode (@modelist)
    {
      if ($mode eq '+' || $mode eq '-') { $flag = $mode }
      if ($mode eq 'o')
      {
        $user = checkhost($cnicks[$i], $nhosts{lc($cnicks[$i])});

        # topiclock set
        if ($flag eq '+' && $user eq $nick && $topiclock{lc($uchan)} eq true)
        {
          sockwrite($sock, "TOPIC $uchan");
        }
        if ($flag eq '+')
        {

          # bitch mode
          if ($user ne $nick && (($bitchmode{lc($uchan)} eq true && ($uflags{lc($user)} !~ /o/ || $user eq false)) || $uflags{lc($user)} =~ /d/))
          {
            sockwrite($sock, "MODE $uchan -o $cnicks[$i]");
          }
        }
        else
        {

          # op protection
          if ($opprot{lc($uchan)} eq true && $user ne false && $user ne $nick && $uflags{lc($user)} =~ /o/ && $uhost ne $nick)
          {
            sockwrite($sock, "MODE $uchan -o+o $uhost $cnicks[$i]");
          }
        }
        $i++;
      }
      if ($mode =~ /[psmntikl]+/ && $modelock{lc($uchan)} eq true && $uhost ne $nick)
      {
        $sendmodes = true;
      }
    }
    if ($sendmodes eq true) { sockwrite($sock, "MODE $uchan"); }
  }
  elsif ($line =~ /^:(.*) TOPIC ([^: ]+) :(.*)$/)
  {
    $uhost = $1;
    $uchan = $2;
    $topic = $3;
    $uhost =~ /(.*)!(.*)\@(.*)/;
    $unick  = $1;
    $uident = $2;
    $uaddy  = $3;
    console("\[$uchan\] $unick sets topic: $topic\n");
    if ($logging{lc($uchan)} eq true) { chanlog($logfiles{lc($uchan)}, "*** $unick sets topic: $topic\n"); }

    if ($topiclock{lc($uchan)} eq true && $unick ne $nick)
    {
      sockwrite($sock, "TOPIC $uchan :$oldtopic{lc($uchan)}");
    }
  }
  elsif ($line =~ /^:(.*) NICK :(.*)$/)
  {
    $uhost   = $1;
    $newnick = $2;
    $uhost =~ /(.*)!(.*)\@(.*)/;
    $unick = $1;
    console("*** $unick is now known as $newnick\n");
    $user = quotemeta($unick);
    foreach $chan (keys %channels)
    {
      $names{lc($chan)} =~ s/$user/$newnick/;
    }
    $nhosts{lc($newnick)} = $nhosts{lc($unick)};
    delete($nhosts{lc($unick)});
  }
  elsif ($line =~ /^:(.*) NOTICE ([^: ]+) :?(.*)$/)
  {
    $uhost  = $1;
    $uchan  = $2;
    $notice = $3;
    $uhost =~ /(.*)!(.*)\@(.*)/;
    $unick  = $1;
    $uident = $2;
    $uaddy  = $3;
    console("\[$unick\] NOTICE: $notice\n");
    foreach $logfile (values %logfiles) { chanlog($logfile, "[$unick] NOTICE: $notice\n"); }

    if ($notice =~ /^This nickname is registered and protected/)
    {
      $nnick = checkhost(lc($unick), $uhost);
      if ($nnick ne false)
      {
        if ($uflags{$nnick} =~ /n/) { console("*** nickserv identify\n"); sockwrite($sock, "PRIVMSG $unick :IDENTIFY $uchans{$nnick}\n"); }
      }
    }
  }
  elsif ($line =~ /^:(.*) INVITE [^ ]* :?(.*)$/)
  {
    $uhost = $1;
    $chan  = lc($2);
    $uhost =~ /(.*)!(.*)\@(.*)/;
    $unick = $1;
    console("\[$unick\] INVITE: $chan\n");
    if ($needinvchan{$chan} eq true)
    {
      $channels{$chan} =~ /[^,]*[,]?(.*)/;
      sockwrite($sock, "JOIN $chan $1");
      delete($needinvchan{$chan});
    }
  }
  elsif ($line =~ /^:(.*) ([0-9]{3}) ([^ :]*) [:]?(.*)$/)
  {
    $text = $4;
    $nik  = $3;
    if ($2 == "001")
    {
      console("*** auto joining\n");
      foreach $chan (keys %channels)
      {
        $channels{$chan} =~ /([^,]*)[,]?(.*)/;
        $key = $2;
        sockwrite($sock, "JOIN $chan $key");
      }
    }
    elsif ($2 == "302")    # userhost
    {
      $text =~ /^(.*)=[\-\+]?(.*)$/;
      $nhosts{lc($1)} = $2;
      foreach $chan (keys %needinvchan)
      {
        if ($needinvchan{$chan} eq true)
        {
          $needinvite{$chan} =~ /^([^ ]*) ([^ ]*) (.*)$/;
          $user  = $1;
          $nnick = $2;
          $msg   = $3;
          $nuser = checkhost($user, $nhosts{lc($nnick)});
          if ($nuser eq $user)
          {
            privmsg($nnick, $msg);
          }
        }
      }
    }
    elsif ($2 == "433" || $2 == "437")    # nick in use
    {
      $curnick++;
      if ($curnick > $#nicks) { $curnick = 0; }
      $nick = quotemeta($nicks[$curnick]);
      sockwrite($sock, "NICK $nicks[$curnick]");
      console("*** changed nick to $nick\n");
    }
    elsif ($2 == "332")                   # topic
    {
      $text =~ /^(.*) :(.*)$/;
      console("*** topic of $1: $2\n");

      if ($topiclock{lc($1)} eq true)
      {
        if (!defined($oldtopic{lc($1)})) { $oldtopic{lc($1)} = $2; }
        elsif ($oldtopic{lc($1)} ne $2) { sockwrite($sock, "TOPIC $1 :$oldtopic{lc($1)}"); }
      }
    }
    elsif ($2 == "353")                   # names
    {
      $text =~ /^.? ([^ :]*)[ ]?:(.*)$/;
      $nchan = $1;
      $data  = $2;
      console("*** users on $nchan: $data\n");
      $names{lc($nchan)} = $data;

      undef($nhosts);
      @cnicks = split (' ', $data);
      foreach $user (@cnicks)
      {
        $user =~ /^[@\+]?(.*)/;
        sockwrite($sock, "USERHOST $1");
      }
      if ($seenflag eq true)
      {
        $seenflag = false;
        $key      = quotemeta($seenkey);
        if ($data =~ /$key/i)
        {
          privmsg($uchan, "$seennick, $seenkey is on the channel right now, don't you see?!");
        }
        elsif (defined($seen{lc($seenkey)}))
        {
          $seen{lc($seenkey)} =~ /^(.*)\x01(.*)\x01(.*)\x01(.*)$/;
          $ts    = time() - int($1);
          $days  = int($ts / 86400);
          $hours = (int($ts / 3600)) % 24;
          $mins  = (int($ts / 60)) % 60;
          $secs  = (int($ts)) % 60;
          privmsg($uchan, "$unick, I last saw $seenkey $4 from $2, $days days $hours hours $mins mins $secs secs ago [$3]");
        }
        else
        {
          privmsg($uchan, "$unick, I never saw $seenkey");
        }
      }
      if ($data =~ /^[ ]*$nick[ ]*$/i)
      {
        console("*** cycling $nchan to regain ops\n");
        sockwrite($sock, "PART $nchan");
        sockwrite($sock, "JOIN $nchan");
      }
    }
    elsif ($2 == "324")    # Channel modes
    {
      $text =~ /^([^ ]*) \+([^ ]*)[ ]?(.*)$/;
      $nchan = $1;
      $modes = $2;
      @parm  = split (' ', $3);
      $data  = $channels{lc($nchan)};
      $data =~ /([^ ,]*)[ ]?([^,]*)[,]?(.*)/;
      $data  = $1;
      @nparm = split (' ', $2);

      $i         = 0;
      $newpmodes = '+';
      $newparms  = '';
      @mlocks    = split (//, $data);
      foreach $ml (@mlocks)
      {
        if ($modes eq '') { $modes = '.' }   # just to set it to a nonalpha char
        if ($ml !~ /[$modes]/)
        {
          if ($ml ne '.') { $newpmodes .= $ml; }
          if ($ml =~ /l/)
          {
            $newparms .= $nparm[$i] . ' ';
          }
          elsif ($ml =~ /k/)
          {
            $newparms .= $nparm[$i] . ' ';
          }
        }
        if ($ml =~ /l/) { $i++; }
        if ($ml =~ /k/) { $i++; }
      }
      if ($newpmodes eq '+') { $newpmodes = ''; }

      $i         = 0;
      $newnmodes = '-';
      @mlocks    = split (//, $modes);
      foreach $ml (@mlocks)
      {
        if ($data eq '') { $data = '.' }    # just to set it to a nonalpha char
        if ($ml !~ /[$data]/)
        {
          if ($ml ne '.') { $newnmodes .= $ml; }
          if ($ml =~ /l/)
          {
            $i++;
          }
          elsif ($ml =~ /k/)
          {
            $newparms .= $parm[$i] . ' ';
          }
        }
        if ($ml =~ /l/) { $i++; }
        if ($ml =~ /k/) { $i++; }
      }
      if ($newnmodes eq '-') { $newnmodes = ''; }

      $newmodes = $newpmodes . $newnmodes;
      if ($newmodes ne '')
      {
        sockwrite($sock, "MODE $nchan $newpmodes$newnmodes $newparms");
      }
    }
    elsif ($2 eq "473")    # Cannot join channel (+i)
    {
      $text =~ /^([^ :]*)/;
      $chan = lc($1);
      $needinvite{$chan} =~ /^([^ ]*) ([^ ]*) (.*)$/;
      $user  = $1;
      $nnick = $2;
      $needinvchan{$chan} = true;
      sockwrite($sock, "USERHOST $nnick");
    }
    elsif ($2 ne "366" && $2 ne "333" && $2 ne "329" && $2 ne "482")
    {
      console("$text\n");
    }
  }
  else
  {
    console("$line\n");
  }
}
